@@ -59,14 +59,21 @@ cache_size(const GraphicsManager *self) { return disk_cache_total_size(self->dis
 // }}}
 
 
+static inline id_type
+next_id(id_type *counter) {
+    id_type ans = ++(*counter);
+    if (UNLIKELY(ans == 0)) ans = ++(*counter);
+    return ans;
+}
+static const unsigned PARENT_DEPTH_LIMIT = 8;
+
 GraphicsManager*
 grman_alloc(void) {
     GraphicsManager *self = (GraphicsManager *)GraphicsManager_Type.tp_alloc(&GraphicsManager_Type, 0);
-    self->images_capacity = self->capacity = 64;
-    self->images = calloc(self->images_capacity, sizeof(Image));
-    self->render_data = calloc(self->capacity, sizeof(ImageRenderData));
+    self->render_data.capacity = 64;
+    self->render_data.item = calloc(self->render_data.capacity, sizeof(self->render_data.item[0]));
     self->storage_limit = DEFAULT_STORAGE_LIMIT;
-    if (self->images == NULL || self->render_data == NULL) {
+    if (self->render_data.item == NULL) {
         PyErr_NoMemory();
         Py_CLEAR(self); return NULL;
     }
@@ -77,8 +84,14 @@ grman_alloc(void) {
 
 static void
 free_refs_data(Image *img) {
-    free(img->refs); img->refs = NULL;
-    img->refcnt = 0; img->refcap = 0;
+    if (img->refs) {
+        ImageRef *s, *tmp;
+        HASH_ITER(hh, img->refs, s, tmp) {
+            HASH_DEL(img->refs, s);
+            free(s);
+        }
+    }
+    img->refs = NULL;
 }
 
 static void
@@ -90,7 +103,7 @@ free_load_data(LoadData *ld) {
 }
 
 static void
-free_image(GraphicsManager *self, Image *img) {
+free_image_resources(GraphicsManager *self, Image *img) {
     if (img->texture_id) free_texture(&img->texture_id);
     ImageAndFrame key = { .image_id=img->internal_id, .frame_id = img->root_frame.id };
     if (!remove_from_cache(self, key) && PyErr_Occurred()) PyErr_Print();
@@ -106,61 +119,82 @@ free_image(GraphicsManager *self, Image *img) {
     self->used_storage -= img->used_storage;
 }
 
+static void
+free_image(GraphicsManager *self, Image *img) {
+    HASH_DEL(self->images, img);
+    free_image_resources(self, img);
+    free(img);
+}
 
 static void
 dealloc(GraphicsManager* self) {
-    size_t i;
     if (self->images) {
-        for (i = 0; i < self->image_count; i++) free_image(self, self->images + i);
-        free(self->images);
+        Image *img, *tmp;
+        HASH_ITER(hh, self->images, img, tmp) {
+            free_image(self, img);
+        }
+        self->images = NULL;
     }
-    free(self->render_data);
+    free(self->render_data.item);
     Py_CLEAR(self->disk_cache);
     Py_TYPE(self)->tp_free((PyObject*)self);
 }
 
-static id_type internal_id_counter = 1;
+static Image*
+img_by_internal_id(const GraphicsManager *self, id_type id) {
+    Image *ans;
+    HASH_FIND(hh, self->images, &id, sizeof(id), ans);
+    return ans;
+}
 
 static Image*
-img_by_internal_id(GraphicsManager *self, id_type id) {
-    for (size_t i = 0; i < self->image_count; i++) {
-        if (self->images[i].internal_id == id) return self->images + i;
+img_by_client_id(const GraphicsManager *self, uint32_t id) {
+    for (Image *img = self->images; img != NULL; img = img->hh.next) {
+        if (img->client_id == id) return img;
     }
     return NULL;
 }
 
 static Image*
-img_by_client_id(GraphicsManager *self, uint32_t id) {
-    for (size_t i = 0; i < self->image_count; i++) {
-        if (self->images[i].client_id == id) return self->images + i;
-    }
-    return NULL;
-}
-
-static Image*
-img_by_client_number(GraphicsManager *self, uint32_t number) {
+img_by_client_number(const GraphicsManager *self, uint32_t number) {
     // get the newest image with the specified number
-    for (size_t i = self->image_count; i-- > 0; ) {
-        if (self->images[i].client_number == number) return self->images + i;
+    Image *ans = NULL;
+    for (Image *img = self->images; img != NULL; img = img->hh.next) {
+        if (img->client_number == number && (!ans || img->internal_id > ans->internal_id)) ans = img;
+    }
+    return ans;
+}
+
+static ImageRef*
+ref_by_internal_id(const Image *img, id_type id) {
+    ImageRef *ans;
+    HASH_FIND(hh, img->refs, &id, sizeof(id), ans);
+    return ans;
+}
+
+
+static ImageRef*
+ref_by_client_id(const Image *img, uint32_t id) {
+    for (ImageRef *ref = img->refs; ref != NULL; ref = ref->hh.next) {
+        if (ref->client_id == id) {
+            return ref;
+        }
     }
     return NULL;
 }
 
-
 static void
-remove_image(GraphicsManager *self, size_t idx) {
-    assert(idx < self->image_count);
-    free_image(self, self->images + idx);
-    remove_i_from_array(self->images, idx, self->image_count);
+remove_image(GraphicsManager *self, Image *img) {
+    free_image(self, img);
     self->layers_dirty = true;
 }
 
 static void
 remove_images(GraphicsManager *self, bool(*predicate)(Image*), id_type skip_image_internal_id) {
-    for (size_t i = self->image_count; i-- > 0;) {
-        Image *img = self->images + i;
+    Image *img, *tmp;
+    HASH_ITER(hh, self->images, img, tmp) {
         if (img->internal_id != skip_image_internal_id && predicate(img)) {
-            remove_image(self, i);
+            remove_image(self, img);
         }
     }
 }
@@ -170,9 +204,13 @@ remove_images(GraphicsManager *self, bool(*predicate)(Image*), id_type skip_imag
 
 static bool
 trim_predicate(Image *img) {
-    return !img->root_frame_data_loaded || !img->refcnt;
+    return !img->root_frame_data_loaded || !img->refs;
 }
 
+static int
+oldest_img_first(const Image *a, const Image *b) {
+    return a->atime - b->atime;
+}
 
 static void
 apply_storage_quota(GraphicsManager *self, size_t storage_limit, id_type currently_added_image_internal_id) {
@@ -180,13 +218,11 @@ apply_storage_quota(GraphicsManager *self, size_t storage_limit, id_type current
     remove_images(self, trim_predicate, currently_added_image_internal_id);
     if (self->used_storage < storage_limit) return;
 
-#define oldest_last(a, b) ((b)->atime < (a)->atime)
-    QSORT(Image, self->images, self->image_count, oldest_last)
-#undef oldest_last
-    while (self->used_storage > storage_limit && self->image_count > 0) {
-        remove_image(self, self->image_count - 1);
+    HASH_SORT(self->images, oldest_img_first);
+    while (self->used_storage > storage_limit && self->images) {
+        remove_image(self, self->images);
     }
-    if (!self->image_count) self->used_storage = 0;  // sanity check
+    if (!self->images) self->used_storage = 0;  // sanity check
 }
 
 static char command_response[512] = {0};
@@ -294,7 +330,7 @@ inflate_png(LoadData *load_data, uint8_t *buf, size_t bufsz) {
 
 static bool
 add_trim_predicate(Image *img) {
-    return !img->root_frame_data_loaded || (!img->client_id && !img->refcnt);
+    return !img->root_frame_data_loaded || (!img->client_id && !img->refs);
 }
 
 bool
@@ -351,28 +387,26 @@ png_path_to_bitmap(const char* path, uint8_t** data, unsigned int* width, unsign
 static Image*
 find_or_create_image(GraphicsManager *self, uint32_t id, bool *existing) {
     if (id) {
-        for (size_t i = 0; i < self->image_count; i++) {
-            if (self->images[i].client_id == id) {
-                *existing = true;
-                return self->images + i;
-            }
+        Image *img = img_by_client_id(self, id);
+        if (img) {
+            *existing = true;
+            return img;
         }
     }
     *existing = false;
-    ensure_space_for(self, images, Image, self->image_count + 1, images_capacity, 64, true);
-    Image *ans = self->images + self->image_count++;
-    zero_at_ptr(ans);
+    Image *ans = calloc(1, sizeof(Image));
+    ans->internal_id = next_id(&self->image_id_counter);
+    HASH_ADD(hh, self->images, internal_id, sizeof(ans->internal_id), ans);
     return ans;
 }
 
 static uint32_t
 get_free_client_id(const GraphicsManager *self) {
-    if (!self->image_count) return 1;
-    uint32_t *client_ids = malloc(sizeof(uint32_t) * self->image_count);
+    if (!self->images) return 1;
+    uint32_t *client_ids = malloc(sizeof(uint32_t) * HASH_COUNT(self->images));
     size_t count = 0;
-    for (size_t i = 0; i < self->image_count; i++) {
-        Image *q = self->images + i;
-        if (q->client_id) client_ids[count++] = q->client_id;
+    for (Image *img = self->images; img != NULL; img = img->hh.next) {
+        if (img->client_id) client_ids[count++] = img->client_id;
     }
     if (!count) { free(client_ids); return 1; }
 #define int_lt(a, b) ((*a)<(*b))
@@ -569,15 +603,14 @@ handle_add_command(GraphicsManager *self, const GraphicsCommand *g, const uint8_
         remove_images(self, add_trim_predicate, 0);
         img = find_or_create_image(self, iid, &existing);
         if (existing) {
+            free_image_resources(self, img);
             img->root_frame_data_loaded = false;
             img->is_drawn = false;
             img->current_frame_shown_at = 0;
             img->extra_framecnt = 0;
-            free_image(self, img);
             *is_dirty = true;
             self->layers_dirty = true;
         } else {
-            img->internal_id = internal_id_counter++;
             img->client_id = iid;
             img->client_number = g->image_number;
             if (!img->client_id && img->client_number) {
@@ -682,6 +715,22 @@ update_dest_rect(ImageRef *ref, uint32_t num_cols, uint32_t num_rows, CellPixelS
     ref->effective_num_cols = num_cols;
 }
 
+static ImageRef*
+create_ref(Image *img, ImageRef *clone_from) {
+    ImageRef *ans = calloc(1, sizeof(ImageRef));
+    if (!ans) fatal("Out of memory creating ImageRef");
+    if (clone_from) {
+        *ans = *clone_from;
+        memset(&ans->hh, 0, sizeof(ans->hh));
+    }
+    ans->internal_id = next_id(&img->ref_id_counter);
+    HASH_ADD(hh, img->refs, internal_id, sizeof(ans->internal_id), ans);
+    return ans;
+}
+
+static inline bool
+is_cell_image(const ImageRef *self) { return self->virtual_ref_id != 0; }
+
 // Create a real image ref for a virtual image ref (placement) positioned in the
 // given cells. This is used for images positioned using Unicode placeholders.
 //
@@ -713,18 +762,17 @@ Image *grman_put_cell_image(GraphicsManager *self, uint32_t screen_row,
     ImageRef *virt_img_ref = NULL;
     if (placement_id) {
         // Find the placement by the id. It must be a virtual placement.
-        for (size_t i = 0; i < img->refcnt; i++) {
-            if (img->refs[i].is_virtual_ref &&
-                img->refs[i].client_id == placement_id) {
-                virt_img_ref = img->refs + i;
+        for (ImageRef *r = img->refs; r != NULL; r = r->hh.next) {
+            if (r->is_virtual_ref && r->client_id == placement_id) {
+                virt_img_ref = r;
                 break;
             }
         }
     } else {
         // Find the first virtual image placement.
-        for (size_t i = 0; i < img->refcnt; i++) {
-            if (img->refs[i].is_virtual_ref) {
-                virt_img_ref = img->refs + i;
+        for (ImageRef *r = img->refs; r != NULL; r = r->hh.next) {
+            if (r->is_virtual_ref) {
+                virt_img_ref = r;
                 break;
             }
         }
@@ -735,7 +783,7 @@ Image *grman_put_cell_image(GraphicsManager *self, uint32_t screen_row,
     // Create the ref structure on stack first. We will not create a real
     // reference if the image is completely out of bounds.
     ImageRef ref = {0};
-    ref.is_cell_image = true;
+    ref.virtual_ref_id = virt_img_ref->internal_id;
 
     uint32_t img_rows = virt_img_ref->num_rows;
     uint32_t img_columns = virt_img_ref->num_cols;
@@ -841,41 +889,103 @@ Image *grman_put_cell_image(GraphicsManager *self, uint32_t screen_row,
     ref.z_index = -1;
 
     // Create a real ref.
-    ensure_space_for(img, refs, ImageRef, img->refcnt + 1, refcap, 16, true);
-    self->layers_dirty = true;
-    ImageRef *real_ref = img->refs + img->refcnt++;
-    *real_ref = ref;
+    ImageRef *real_ref = create_ref(img, &ref);
+
     img->atime = monotonic();
+    self->layers_dirty = true;
 
     update_src_rect(real_ref, img);
     update_dest_rect(real_ref, ref.num_cols, ref.num_rows, cell);
     return img;
 }
 
+static void remove_ref(Image *img, ImageRef *ref);
+
+static bool
+has_good_ancestry(GraphicsManager *self, ImageRef *ref) {
+    ImageRef *r = ref;
+    unsigned depth = 0;
+    while (r->parent.img) {
+        if (r == ref && depth) {
+            set_command_failed_response("ECYCLE", "This parent reference creates a cycle");
+            return false;
+        }
+        if (depth++ >= PARENT_DEPTH_LIMIT) {
+            set_command_failed_response("ETOODEEP", "Too many levels of parent references");
+            return false;
+        }
+        Image *parent = img_by_internal_id(self, r->parent.img);
+        if (!parent) {
+            set_command_failed_response("ENOENT", "One of the ancestors of this ref with image id: %llu not found", r->parent.img);
+            return false;
+        }
+        ImageRef *parent_ref = ref_by_internal_id(parent, r->parent.ref);
+        if (!parent_ref) {
+            set_command_failed_response("ENOENT", "One of the ancestors of this ref with image id: %llu and ref id: %llu not found", r->parent.img, r->parent.ref);
+            return false;
+        }
+        r = parent_ref;
+    }
+    return true;
+}
+
 static uint32_t
 handle_put_command(GraphicsManager *self, const GraphicsCommand *g, Cursor *c, bool *is_dirty, Image *img, CellPixelSize cell) {
+    if (g->unicode_placement && g->parent_id) {
+        set_command_failed_response("EINVAL", "Put command creating a virtual placement cannot refer to a parent"); return g->id;
+    }
     if (img == NULL) {
         if (g->id) img = img_by_client_id(self, g->id);
         else if (g->image_number) img = img_by_client_number(self, g->image_number);
         if (img == NULL) { set_command_failed_response("ENOENT", "Put command refers to non-existent image with id: %u and number: %u", g->id, g->image_number); return g->id; }
     }
     if (!img->root_frame_data_loaded) { set_command_failed_response("ENOENT", "Put command refers to image with id: %u that could not load its data", g->id); return img->client_id; }
-    ensure_space_for(img, refs, ImageRef, img->refcnt + 1, refcap, 16, true);
-    *is_dirty = true;
-    self->layers_dirty = true;
+    id_type parent_id = 0, parent_placement_id = 0;
+    if (g->parent_id) {
+        Image *parent = img_by_client_id(self, g->parent_id);
+        if (!parent) {
+            set_command_failed_response("ENOPARENT", "Put command refers to a parent image with id: %u that does not exist", g->parent_id);
+            return g->id;
+        }
+        if (!parent->refs) {
+            set_command_failed_response("ENOPARENT", "Put command refers to a parent image with id: %u that has no placements", g->parent_id);
+            return g->id;
+        }
+        ImageRef *parent_ref = parent->refs;
+        if (g->parent_placement_id) {
+            parent_ref = ref_by_client_id(parent, g->parent_placement_id);
+            if (!parent_ref) {
+                set_command_failed_response("ENOPARENT", "Put command refers to a parent image placement with id: %u and placement id: %u that does not exist", g->parent_id, g->parent_placement_id);
+                return g->id;
+            }
+        }
+        parent_id = parent->internal_id;
+        parent_placement_id = parent_ref->internal_id;
+    }
     ImageRef *ref = NULL;
     if (g->placement_id && img->client_id) {
-        for (size_t i=0; i < img->refcnt; i++) {
-            if (img->refs[i].client_id == g->placement_id) {
-                ref = img->refs + i;
+        for (ImageRef *r = img->refs; r != NULL; r = r->hh.next) {
+            if (r->client_id == g->placement_id) {
+                ref = r;
+                if (parent_id && parent_id == img->internal_id && parent_placement_id && parent_placement_id == r->internal_id) {
+                    set_command_failed_response("EINVAL", "Put command refers to itself as its own parent");
+                    return g->id;
+                }
+                if (parent_id && parent_placement_id) {
+                    id_type rp = ref->parent.img, rpp = ref->parent.ref;
+                    ref->parent.img = parent_id; ref->parent.ref = parent_placement_id;
+                    bool ok = has_good_ancestry(self, ref);
+                    ref->parent.img = rp; ref->parent.ref = rpp;
+                    if (!ok) return g->id;
+                }
                 break;
             }
         }
     }
-    if (ref == NULL) {
-        ref = img->refs + img->refcnt++;
-        zero_at_ptr(ref);
-    }
+    if (ref == NULL) ref = create_ref(img, NULL);
+
+    *is_dirty = true;
+    self->layers_dirty = true;
     img->atime = monotonic();
     ref->src_x = g->x_offset; ref->src_y = g->y_offset; ref->src_width = g->width ? g->width : img->width; ref->src_height = g->height ? g->height : img->height;
     ref->src_width = MIN(ref->src_width, img->width - ((float)img->width > ref->src_x ? ref->src_x : (float)img->width));
@@ -888,13 +998,25 @@ handle_put_command(GraphicsManager *self, const GraphicsCommand *g, Cursor *c, b
     if (img->client_id) ref->client_id = g->placement_id;
     update_src_rect(ref, img);
     update_dest_rect(ref, g->num_cells, g->num_lines, cell);
+    ref->parent.img = parent_id;
+    ref->parent.ref = parent_placement_id;
+    ref->parent.offset.x = g->offset_from_parent_x;
+    ref->parent.offset.y = g->offset_from_parent_y;
+    ref->is_virtual_ref = false;
     if (g->unicode_placement) {
         ref->is_virtual_ref = true;
         ref->start_row = ref->start_column = 0;
     }
-    // Move the cursor, the screen will take care of ensuring it is in bounds
-    if (g->cursor_movement != 1 && !g->unicode_placement) {
-        c->x += ref->effective_num_cols; c->y += ref->effective_num_rows - 1;
+    if (ref->parent.img) {
+        if (!has_good_ancestry(self, ref)) {
+            remove_ref(img, ref);
+            return g->id;
+        }
+    } else {
+        // Move the cursor, the screen will take care of ensuring it is in bounds
+        if (g->cursor_movement != 1 && !g->unicode_placement) {
+            c->x += ref->effective_num_cols; c->y += ref->effective_num_rows - 1;
+        }
     }
     return img->client_id;
 }
@@ -919,17 +1041,59 @@ gpu_data_for_image(ImageRenderData *ans, float left, float top, float right, flo
     ans->group_count = 1;
 }
 
+static bool
+resolve_cell_ref(const Image *img, id_type virt_ref_id, int32_t *start_row, int32_t *start_column) {
+    *start_row = 0; *start_column = 0;
+    bool found = false;
+    for (ImageRef *ref = img->refs; ref != NULL; ref = ref->hh.next) {
+        if (ref->virtual_ref_id == virt_ref_id) {
+            if (!found || ref->start_row < *start_row) *start_row = ref->start_row;
+            if (!found || ref->start_column < *start_column) *start_column = ref->start_column;
+            found = true;
+
+        }
+    }
+    return found;
+}
+
+static bool
+resolve_parent_offset(const GraphicsManager *self, const ImageRef *ref, int32_t *start_row, int32_t *start_column, bool *is_virtual_ref) {
+    *start_row = 0; *start_column = 0;
+    int32_t x = 0, y = 0;
+    unsigned depth = 0;
+    ImageRef cell_ref = {0};
+    while (ref->parent.img) {
+        if (depth++ >= PARENT_DEPTH_LIMIT) return false;  // either a cycle or too many ancestors
+        Image *img = img_by_internal_id(self, ref->parent.img);
+        if (!img) return false;
+        ImageRef *parent = ref_by_internal_id(img, ref->parent.ref);
+        if (!parent) return false;
+        if (parent->is_virtual_ref) {
+            *is_virtual_ref = true;
+            if (!resolve_cell_ref(img, parent->internal_id, &cell_ref.start_row, &cell_ref.start_column)) return false;
+            parent = &cell_ref;
+        }
+        x += ref->parent.offset.x;
+        y += ref->parent.offset.y;
+        ref = parent;
+    }
+    *start_row = ref->start_row + y;
+    *start_column = ref->start_column + x;
+    return true;
+}
+
+
 bool
 grman_update_layers(GraphicsManager *self, unsigned int scrolled_by, float screen_left, float screen_top, float dx, float dy, unsigned int num_cols, unsigned int num_rows, CellPixelSize cell) {
     if (self->last_scrolled_by != scrolled_by) self->layers_dirty = true;
     self->last_scrolled_by = scrolled_by;
     if (!self->layers_dirty) return false;
     self->layers_dirty = false;
-    size_t i, j;
+    size_t i;
     self->num_of_below_refs = 0;
     self->num_of_negative_refs = 0;
     self->num_of_positive_refs = 0;
-    Image *img; ImageRef *ref;
+    Image *img, *tmpimg; ImageRef *ref, *tmpref;
     ImageRect r;
     float screen_width = dx * num_cols, screen_height = dy * num_rows;
     float screen_bottom = screen_top - screen_height;
@@ -938,21 +1102,33 @@ grman_update_layers(GraphicsManager *self, unsigned int scrolled_by, float scree
     float y0 = screen_top - dy * scrolled_by;
 
     // Iterate over all visible refs and create render data
-    self->count = 0;
-    for (i = 0; i < self->image_count; i++) {
-        img = self->images + i;
+    self->render_data.count = 0;
+
+    HASH_ITER(hh, self->images, img, tmpimg) {
         bool was_drawn = img->is_drawn;
+        bool ref_removed = false;
         img->is_drawn = false;
 
-        for (j = 0; j < img->refcnt; j++) { ref = img->refs + j;
+        HASH_ITER(hh, img->refs, ref, tmpref) {
             if (ref->is_virtual_ref) continue;
-            r.top = y0 - ref->start_row * dy - dy * (float)ref->cell_y_offset / (float)cell.height;
-            if (ref->num_rows > 0) r.bottom = y0 - (ref->start_row + (int32_t)ref->num_rows) * dy;
+            int32_t start_row = ref->start_row, start_column = ref->start_column;
+            if (ref->parent.img) {
+                bool is_virtual_ref;
+                if (!resolve_parent_offset(self, ref, &start_row, &start_column, &is_virtual_ref)) {
+                    if (!is_virtual_ref) {
+                        remove_ref(img, ref);
+                        ref_removed = true;
+                    }
+                    continue;
+                }
+            }
+            r.top = y0 - start_row * dy - dy * (float)ref->cell_y_offset / (float)cell.height;
+            if (ref->num_rows > 0) r.bottom = y0 - (start_row + (int32_t)ref->num_rows) * dy;
             else r.bottom = r.top - screen_height * (float)ref->src_height / screen_height_px;
             if (r.top <= screen_bottom || r.bottom >= screen_top) continue;  // not visible
 
-            r.left = screen_left + ref->start_column * dx + dx * (float)ref->cell_x_offset / (float) cell.width;
-            if (ref->num_cols > 0) r.right = screen_left + (ref->start_column + (int32_t)ref->num_cols) * dx;
+            r.left = screen_left + start_column * dx + dx * (float)ref->cell_x_offset / (float) cell.width;
+            if (ref->num_cols > 0) r.right = screen_left + (start_column + (int32_t)ref->num_cols) * dx;
             else r.right = r.left + screen_width * (float)ref->src_width / screen_width_px;
 
             if (ref->z_index < ((int32_t)INT32_MIN/2))
@@ -961,35 +1137,39 @@ grman_update_layers(GraphicsManager *self, unsigned int scrolled_by, float scree
                 self->num_of_negative_refs++;
             else
                 self->num_of_positive_refs++;
-            ensure_space_for(self, render_data, ImageRenderData, self->count + 1, capacity, 64, true);
-            ImageRenderData *rd = self->render_data + self->count;
+            ensure_space_for(&(self->render_data), item, ImageRenderData, self->render_data.count + 1, capacity, 64, true);
+            ImageRenderData *rd = self->render_data.item + self->render_data.count;
             zero_at_ptr(rd);
             rd->dest_rect = r; rd->src_rect = ref->src_rect;
-            self->count++;
-            rd->z_index = ref->z_index; rd->image_id = img->internal_id;
+            self->render_data.count++;
+            rd->z_index = ref->z_index; rd->image_id = img->internal_id; rd->ref_id = ref->internal_id;
             rd->texture_id = img->texture_id;
             img->is_drawn = true;
         }
+        if (ref_removed && !img->refs) {
+            remove_image(self, img);
+            continue;
+        }
         if (img->is_drawn && !was_drawn && img->animation_state != ANIMATION_STOPPED && img->extra_framecnt && img->animation_duration) {
             self->has_images_needing_animation = true;
             global_state.check_for_active_animated_images = true;
         }
     }
-    if (!self->count) return false;
+    if (!self->render_data.count) return false;
     // Sort visible refs in draw order (z-index, img)
 #define lt(a, b) ( (a)->z_index < (b)->z_index || ((a)->z_index == (b)->z_index && (a)->image_id < (b)->image_id) )
-    QSORT(ImageRenderData, self->render_data, self->count, lt);
+    QSORT(ImageRenderData, self->render_data.item, self->render_data.count, lt);
 #undef lt
     // Calculate the group counts
     i = 0;
-    while (i < self->count) {
-        id_type num_identical = 1, image_id = self->render_data[i].image_id, start = i;
-        while (++i < self->count) {
-            if (self->render_data[i].image_id != image_id) break;
+    while (i < self->render_data.count) {
+        id_type num_identical = 1, image_id = self->render_data.item[i].image_id, start = i;
+        while (++i < self->render_data.count) {
+            if (self->render_data.item[i].image_id != image_id) break;
             num_identical++;
         }
         while (num_identical > 0) {
-            self->render_data[start++].group_count = num_identical--;
+            self->render_data.item[start++].group_count = num_identical--;
         }
     }
     return true;
@@ -1466,8 +1646,7 @@ scan_active_animations(GraphicsManager *self, const monotonic_t now, monotonic_t
     if (!self->has_images_needing_animation) return dirtied;
     self->has_images_needing_animation = false;
     self->context_made_current_for_this_command = os_window_context_set;
-    for (size_t i = self->image_count; i-- > 0;) {
-        Image *img = self->images + i;
+    for (Image *img = self->images; img != NULL; img = img->hh.next) {
         if (image_is_animatable(img)) {
             Frame *f = current_frame(img);
             if (f) {
@@ -1566,20 +1745,28 @@ handle_compose_command(GraphicsManager *self, bool *is_dirty, const GraphicsComm
 
 // Image lifetime/scrolling {{{
 
+static void
+remove_ref(Image *img, ImageRef *ref) {
+    HASH_DEL(img->refs, ref);
+    free(ref);
+}
+
 static void
 filter_refs(GraphicsManager *self, const void* data, bool free_images, bool (*filter_func)(const ImageRef*, Image*, const void*, CellPixelSize), CellPixelSize cell, bool only_first_image) {
     bool matched = false;
-    for (size_t i = self->image_count; i-- > 0;) {
-        Image *img = self->images + i;
-        for (size_t j = img->refcnt; j-- > 0;) {
-            ImageRef *ref = img->refs + j;
-            if (filter_func(ref, img, data, cell)) {
-                remove_i_from_array(img->refs, j, img->refcnt);
-                self->layers_dirty = true;
-                matched = true;
+    Image *img, *tmp;
+    HASH_ITER(hh, self->images, img, tmp) {
+        if (img->refs) {
+            ImageRef *ref, *tmp;
+            HASH_ITER(hh, img->refs, ref, tmp) {
+                if (filter_func(ref, img, data, cell)) {
+                    remove_ref(img, ref);
+                    self->layers_dirty = true;
+                    matched = true;
+                }
             }
         }
-        if (img->refcnt == 0 && (free_images || img->client_id == 0)) remove_image(self, i);
+        if (!img->refs && (free_images || img->client_id == 0)) remove_image(self, img);
         if (only_first_image && matched) break;
     }
 }
@@ -1587,15 +1774,18 @@ filter_refs(GraphicsManager *self, const void* data, bool free_images, bool (*fi
 
 static void
 modify_refs(GraphicsManager *self, const void* data, bool (*filter_func)(ImageRef*, Image*, const void*, CellPixelSize), CellPixelSize cell) {
-    for (size_t i = self->image_count; i-- > 0;) {
-        Image *img = self->images + i;
-        for (size_t j = img->refcnt; j-- > 0;) {
-            if (filter_func(img->refs + j, img, data, cell)) remove_i_from_array(img->refs, j, img->refcnt);
+    Image *img, *tmp;
+    HASH_ITER(hh, self->images, img, tmp) {
+        if (img->refs) {
+            ImageRef *ref, *tmp;
+            HASH_ITER(hh, img->refs, ref, tmp) {
+                if (filter_func(ref, img, data, cell)) remove_ref(img, ref);
+            }
         }
-        if (img->refcnt == 0 && img->client_id == 0 && img->client_number == 0) {
+        if (!img->refs && img->client_id == 0 && img->client_number == 0) {
             // references have all scrolled off the history buffer and the image has no way to reference it
             // to create new references so remove it.
-            remove_image(self, i);
+            remove_image(self, img);
         }
     }
 }
@@ -1653,7 +1843,7 @@ scroll_filter_margins_func(ImageRef* ref, Image* img, const void* data, CellPixe
 
 void
 grman_scroll_images(GraphicsManager *self, const ScrollData *data, CellPixelSize cell) {
-    if (self->image_count) {
+    if (self->images) {
         self->layers_dirty = true;
         modify_refs(self, data, data->has_margins ? scroll_filter_margins_func : scroll_filter_func, cell);
     }
@@ -1661,7 +1851,7 @@ grman_scroll_images(GraphicsManager *self, const ScrollData *data, CellPixelSize
 
 static bool
 cell_image_row_filter_func(const ImageRef *ref, Image UNUSED *img, const void *data, CellPixelSize cell UNUSED) {
-    if (ref->is_virtual_ref || !ref->is_cell_image)
+    if (ref->is_virtual_ref || !is_cell_image(ref))
         return false;
     int32_t top = *(int32_t *)data;
     int32_t bottom = *((int32_t *)data + 1);
@@ -1670,7 +1860,7 @@ cell_image_row_filter_func(const ImageRef *ref, Image UNUSED *img, const void *d
 
 static bool
 cell_image_filter_func(const ImageRef *ref, Image UNUSED *img, const void *data UNUSED, CellPixelSize cell UNUSED) {
-    return !ref->is_virtual_ref && ref->is_cell_image;
+    return !ref->is_virtual_ref && is_cell_image(ref);
 }
 
 // Remove cell images within the given region.
@@ -1696,7 +1886,7 @@ clear_filter_func(const ImageRef *ref, Image UNUSED *img, const void UNUSED *dat
 
 static bool
 clear_filter_func_noncell(const ImageRef *ref, Image UNUSED *img, const void UNUSED *data, CellPixelSize cell UNUSED) {
-    if (ref->is_virtual_ref || ref->is_cell_image) return false;
+    if (ref->is_virtual_ref || is_cell_image(ref)) return false;
     return ref->start_row + (int32_t)ref->effective_num_rows > 0;
 }
 
@@ -1728,21 +1918,21 @@ number_filter_func(const ImageRef *ref, Image *img, const void *data, CellPixelS
 
 static bool
 x_filter_func(const ImageRef *ref, Image UNUSED *img, const void *data, CellPixelSize cell UNUSED) {
-    if (ref->is_virtual_ref || ref->is_cell_image) return false;
+    if (ref->is_virtual_ref || is_cell_image(ref)) return false;
     const GraphicsCommand *g = data;
     return ref->start_column <= (int32_t)g->x_offset - 1 && ((int32_t)g->x_offset - 1) < ((int32_t)(ref->start_column + ref->effective_num_cols));
 }
 
 static bool
 y_filter_func(const ImageRef *ref, Image UNUSED *img, const void *data, CellPixelSize cell UNUSED) {
-    if (ref->is_virtual_ref || ref->is_cell_image) return false;
+    if (ref->is_virtual_ref || is_cell_image(ref)) return false;
     const GraphicsCommand *g = data;
     return ref->start_row <= (int32_t)g->y_offset - 1 && ((int32_t)g->y_offset - 1) < ((int32_t)(ref->start_row + ref->effective_num_rows));
 }
 
 static bool
 z_filter_func(const ImageRef *ref, Image UNUSED *img, const void *data, CellPixelSize cell UNUSED) {
-    if (ref->is_virtual_ref || ref->is_cell_image) return false;
+    if (ref->is_virtual_ref || is_cell_image(ref)) return false;
     const GraphicsCommand *g = data;
     return ref->z_index == g->z_index;
 }
@@ -1750,13 +1940,13 @@ z_filter_func(const ImageRef *ref, Image UNUSED *img, const void *data, CellPixe
 
 static bool
 point_filter_func(const ImageRef *ref, Image *img, const void *data, CellPixelSize cell) {
-    if (ref->is_virtual_ref || ref->is_cell_image) return false;
+    if (ref->is_virtual_ref || is_cell_image(ref)) return false;
     return x_filter_func(ref, img, data, cell) && y_filter_func(ref, img, data, cell);
 }
 
 static bool
 point3d_filter_func(const ImageRef *ref, Image *img, const void *data, CellPixelSize cell) {
-    if (ref->is_virtual_ref || ref->is_cell_image) return false;
+    if (ref->is_virtual_ref || is_cell_image(ref)) return false;
     return z_filter_func(ref, img, data, cell) && point_filter_func(ref, img, data, cell);
 }
 
@@ -1799,7 +1989,7 @@ handle_delete_command(GraphicsManager *self, const GraphicsCommand *g, Cursor *c
 #undef D
 #undef I
     }
-    if (!self->image_count && self->count) self->count = 0;
+    if (!self->images && self->render_data.count) self->render_data.count = 0;
 }
 
 // }}}
@@ -1810,11 +2000,9 @@ grman_resize(GraphicsManager *self, index_type old_lines UNUSED, index_type line
     self->layers_dirty = true;
     if (columns == old_columns && num_content_lines_before > num_content_lines_after) {
         const unsigned int vertical_shrink_size = num_content_lines_before - num_content_lines_after;
-        for (size_t i = self->image_count; i-- > 0;) {
-            img = self->images + i;
-            for (size_t j = img->refcnt; j-- > 0;) {
-                ref = img->refs + j;
-                if (ref->is_virtual_ref || ref->is_cell_image) continue;
+        for (img = self->images; img != NULL; img = img->hh.next) {
+            for (ref = img->refs; ref != NULL; ref = ref->hh.next) {
+                if (ref->is_virtual_ref || is_cell_image(ref)) continue;
                 ref->start_row -= vertical_shrink_size;
             }
         }
@@ -1825,11 +2013,9 @@ void
 grman_rescale(GraphicsManager *self, CellPixelSize cell) {
     ImageRef *ref; Image *img;
     self->layers_dirty = true;
-    for (size_t i = self->image_count; i-- > 0;) {
-        img = self->images + i;
-        for (size_t j = img->refcnt; j-- > 0;) {
-            ref = img->refs + j;
-            if (ref->is_virtual_ref || ref->is_cell_image) continue;
+    for (img = self->images; img != NULL; img = img->hh.next) {
+        for (ref = img->refs; ref != NULL; ref = ref->hh.next) {
+            if (ref->is_virtual_ref || is_cell_image(ref)) continue;
             ref->cell_x_offset = MIN(ref->cell_x_offset, cell.width - 1);
             ref->cell_y_offset = MIN(ref->cell_y_offset, cell.height - 1);
             update_dest_rect(ref, ref->num_cols, ref->num_rows, cell);
@@ -1958,7 +2144,7 @@ image_as_dict(GraphicsManager *self, Image *img) {
     CoalescedFrameData cfd = get_coalesced_frame_data(self, img, &img->root_frame);
     if (!cfd.buf) { PyErr_SetString(PyExc_RuntimeError, "Failed to get data for root frame"); return NULL; }
     PyObject *ans = Py_BuildValue("{sI sI sI sI sI sI sI " "sO sI sO " "sI sI sI " "sI sy# sN}",
-        U(texture_id), U(client_id), U(width), U(height), U(internal_id), U(refcnt), U(client_number),
+        U(texture_id), U(client_id), U(width), U(height), U(internal_id), "refs.count", (unsigned int)HASH_COUNT(img->refs), U(client_number),
 
         B(root_frame_data_loaded), U(animation_state), "is_4byte_aligned", img->root_frame.is_4byte_aligned ? Py_True : Py_False,
 
@@ -2019,12 +2205,12 @@ W(update_layers) {
     CellPixelSize cell;
     PA("IffffIIII", &scrolled_by, &xstart, &ystart, &dx, &dy, &sx, &sy, &cell.width, &cell.height);
     grman_update_layers(self, scrolled_by, xstart, ystart, dx, dy, sx, sy, cell);
-    PyObject *ans = PyTuple_New(self->count);
-    for (size_t i = 0; i < self->count; i++) {
-        ImageRenderData *r = self->render_data + i;
+    PyObject *ans = PyTuple_New(self->render_data.count);
+    for (size_t i = 0; i < self->render_data.count; i++) {
+        ImageRenderData *r = self->render_data.item + i;
 #define R(which) Py_BuildValue("{sf sf sf sf}", "left", r->which.left, "top", r->which.top, "right", r->which.right, "bottom", r->which.bottom)
         PyTuple_SET_ITEM(ans, i,
-            Py_BuildValue("{sN sN sI si sK}", "src_rect", R(src_rect), "dest_rect", R(dest_rect), "group_count", r->group_count, "z_index", r->z_index, "image_id", r->image_id)
+            Py_BuildValue("{sN sN sI si sK sK}", "src_rect", R(src_rect), "dest_rect", R(dest_rect), "group_count", r->group_count, "z_index", r->z_index, "image_id", r->image_id, "ref_id", r->ref_id)
         );
 #undef R
     }
@@ -2040,8 +2226,17 @@ static PyMethodDef methods[] = {
     {NULL}  /* Sentinel */
 };
 
+static PyObject*
+get_image_count(GraphicsManager *self, void* closure UNUSED) {
+    unsigned long ans = HASH_COUNT(self->images);
+    return PyLong_FromUnsignedLong(ans);
+}
+static PyGetSetDef getsets[] = {
+    {"image_count", (getter)get_image_count, NULL, NULL, NULL},
+    {NULL},
+};
+
 static PyMemberDef members[] = {
-    {"image_count", T_PYSSIZET, offsetof(GraphicsManager, image_count), READONLY, "image_count"},
     {"storage_limit", T_PYSSIZET, offsetof(GraphicsManager, storage_limit), 0, "storage_limit"},
     {"disk_cache", T_OBJECT_EX, offsetof(GraphicsManager, disk_cache), READONLY, "disk_cache"},
     {NULL},
@@ -2057,6 +2252,7 @@ PyTypeObject GraphicsManager_Type = {
     .tp_new = new,
     .tp_methods = methods,
     .tp_members = members,
+    .tp_getset = getsets,
 };
 
 static PyObject*
