@@ -7,10 +7,11 @@
 #pragma once
 #include "data-types.h"
 #include "monotonic.h"
+#include "kitty-uthash.h"
 
 typedef struct {
     unsigned char action, transmission_type, compressed, delete_action;
-    uint32_t format, more, id, image_number, data_sz, data_offset, placement_id, quiet;
+    uint32_t format, more, id, image_number, data_sz, data_offset, placement_id, quiet, parent_id, parent_placement_id;
     uint32_t width, height, x_offset, y_offset;
     union { uint32_t cursor_movement, compose_mode; };
     union { uint32_t cell_x_offset, blend_mode; };
@@ -22,6 +23,7 @@ typedef struct {
     union { int32_t z_index, gap; };
     size_t payload_sz;
     bool unicode_placement;
+    int32_t offset_from_parent_x, offset_from_parent_y;
 } GraphicsCommand;
 
 typedef struct {
@@ -35,12 +37,21 @@ typedef struct {
     int32_t start_row, start_column;
     uint32_t client_id;
     ImageRect src_rect;
-    // Indicates whether this reference represents a cell image that should be
+    // Indicates whether this reference represents a cell ref that should be
     // removed when the corresponding cells are modified.
-    bool is_cell_image;
+    // The internal id of the virtual ref this cell image was created from. Is a cell ref if this is non-zero.
+    id_type virtual_ref_id;
     // Virtual refs are not displayed but they can be used as prototypes for
     // refs placed using unicode placeholders.
     bool is_virtual_ref;
+
+    struct {
+        id_type img, ref;
+        struct { int32_t x, y; } offset;
+    } parent;
+
+    id_type internal_id;
+    hash_handle_type hh;
 } ImageRef;
 
 typedef struct {
@@ -56,16 +67,19 @@ typedef struct {
 
     bool root_frame_data_loaded;
     ImageRef *refs;
+    id_type ref_id_counter;
     Frame *extra_frames, root_frame;
     uint32_t current_frame_index, frame_id_counter;
     uint64_t animation_duration;
-    size_t refcnt, refcap, extra_framecnt;
+    size_t extra_framecnt;
     monotonic_t atime;
     size_t used_storage;
     bool is_drawn;
     AnimationState animation_state;
     uint32_t max_loops, current_loop;
     monotonic_t current_frame_shown_at;
+
+    hash_handle_type hh;
 } Image;
 
 typedef struct {
@@ -79,7 +93,7 @@ typedef struct {
     ImageRect src_rect, dest_rect;
     uint32_t texture_id, group_count;
     int z_index;
-    id_type image_id;
+    id_type image_id, ref_id;
 } ImageRenderData;
 
 typedef struct {
@@ -106,11 +120,14 @@ typedef struct {
 typedef struct {
     PyObject_HEAD
 
-    size_t image_count, images_capacity, storage_limit;
+    size_t storage_limit;
     LoadData currently_loading;
     Image *images;
-    size_t count, capacity;
-    ImageRenderData *render_data;
+    id_type image_id_counter;
+    struct {
+        size_t count, capacity;
+        ImageRenderData *item;
+    } render_data;
     bool layers_dirty;
     // The number of images below MIN_ZINDEX / 2, then the number of refs between MIN_ZINDEX / 2 and -1 inclusive, then the number of refs above 0 inclusive.
     size_t num_of_below_refs, num_of_negative_refs, num_of_positive_refs;
