@@ -22,6 +22,13 @@ typedef struct {
 } UrlPrefix;
 
 typedef enum AdjustmentUnit { POINT = 0, PERCENT = 1, PIXEL = 2 } AdjustmentUnit;
+typedef enum UnderlineHyperlinks { UNDERLINE_ON_HOVER = 0, UNDERLINE_ALWAYS = 1, UNDERLINE_NEVER = 2 } UnderlineHyperlinks;
+
+struct MenuItem {
+    const char* *location;
+    size_t location_count;
+    const char *definition;
+};
 
 typedef struct {
     monotonic_t visual_bell_duration, cursor_blink_interval, cursor_stop_blinking_after, mouse_hide_wait, click_interval;
@@ -89,9 +96,11 @@ typedef struct {
         float val; AdjustmentUnit unit;
     } underline_position, underline_thickness, strikethrough_position, strikethrough_thickness, cell_width, cell_height, baseline;
     bool show_hyperlink_targets;
+    UnderlineHyperlinks underline_hyperlinks;
     int background_blur;
     long macos_titlebar_color;
     unsigned long wayland_titlebar_color;
+    struct { struct MenuItem *entries; size_t count; } global_menu;
 } Options;
 
 typedef struct WindowLogoRenderData {
@@ -343,6 +352,7 @@ typedef enum {
     HIDE_OTHERS,
     MINIMIZE,
     QUIT,
+    USER_MENU_ACTION,
 
     NUM_COCOA_PENDING_ACTIONS
 } CocoaPendingAction;
@@ -381,3 +391,4 @@ void set_ignore_os_keyboard_processing(bool enabled);
 void update_menu_bar_title(PyObject *title UNUSED);
 void change_live_resize_state(OSWindow*, bool);
 bool render_os_window(OSWindow *w, monotonic_t now, bool ignore_render_frames, bool scan_for_animated_images);
+void update_mouse_pointer_shape(void);
