@@ -19,6 +19,7 @@ import (
 	"kitty/tools/cli"
 	"kitty/tools/cmd/at"
 	"kitty/tools/cmd/edit_in_kitty"
+	"kitty/tools/cmd/mouse_demo"
 	"kitty/tools/cmd/pytest"
 	"kitty/tools/cmd/run_shell"
 	"kitty/tools/cmd/show_error"
@@ -50,6 +51,15 @@ func KittyToolEntryPoints(root *cli.Command) {
 	unicode_input.EntryPoint(root)
 	// show_key
 	show_key.EntryPoint(root)
+	// mouse_demo
+	root.AddSubCommand(&cli.Command{
+		Name:             "mouse-demo",
+		ShortDescription: "Demo the mouse handling kitty implements for terminal programs",
+		OnlyArgsAllowed:  true,
+		Run: func(cmd *cli.Command, args []string) (rc int, err error) {
+			return mouse_demo.Run(args)
+		},
+	})
 	// hyperlinked_grep
 	hyperlinked_grep.EntryPoint(root)
 	// ask
